import os

import django
from celery import Celery


DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:",}}


SITE_ID = 1

MIDDLEWARE = (
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "defender.middleware.FailedLoginMiddleware",
)

ROOT_URLCONF = "defender.test_urls"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.admin",
    "defender",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.request",
            ],
        },
    },
]

SECRET_KEY = os.environ.get("SECRET_KEY", "too-secret-for-test")

LOGIN_REDIRECT_URL = "/admin"

DEFENDER_LOGIN_FAILURE_LIMIT = 10
DEFENDER_COOLOFF_TIME = 2
DEFENDER_REDIS_URL = None
# use mock redis in unit tests locally.
DEFENDER_MOCK_REDIS = True

# celery settings
CELERY_ALWAYS_EAGER = True
BROKER_BACKEND = "memory"
BROKER_URL = "memory://"

# set the default Django settings module for the 'celery' program.
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "defender.test_settings")

app = Celery("defender")

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object("django.conf:settings")
app.autodiscover_tasks(lambda: INSTALLED_APPS)



from django.contrib.auth.views import LoginView
from django.utils.decorators import method_decorator

from .decorators import watch_login


class FailedLoginMiddleware:
    """ Failed login middleware """

    patched = False

    def __init__(self, get_response):
        self.get_response = get_response

        # Watch the auth login.
        # Monkey-patch only once - otherwise we would be recording
        # failed attempts multiple times!
        if not FailedLoginMiddleware.patched:
            our_decorator = watch_login()
            watch_login_method = method_decorator(our_decorator)
            LoginView.dispatch = watch_login_method(LoginView.dispatch)

            FailedLoginMiddleware.patched = True

    def __call__(self, request):
        response = self.get_response(request)
        return response


from django.conf import settings
from django.utils.translation import gettext_lazy


def get_setting(variable, default=None):
    """ get the 'variable' from settings if not there use the
    provided default """
    return getattr(settings, variable, default)


# redis server host
DEFENDER_REDIS_URL = get_setting("DEFENDER_REDIS_URL")

# redis password quote for special character
DEFENDER_REDIS_PASSWORD_QUOTE = get_setting("DEFENDER_REDIS_PASSWORD_QUOTE", False)

# reuse declared cache from django settings
DEFENDER_REDIS_NAME = get_setting("DEFENDER_REDIS_NAME")

MOCK_REDIS = get_setting("DEFENDER_MOCK_REDIS", False)

# see if the user has overridden the failure limit
FAILURE_LIMIT = get_setting("DEFENDER_LOGIN_FAILURE_LIMIT", 3)
USERNAME_FAILURE_LIMIT = get_setting(
    "DEFENDER_LOGIN_FAILURE_LIMIT_USERNAME", FAILURE_LIMIT
)
IP_FAILURE_LIMIT = get_setting("DEFENDER_LOGIN_FAILURE_LIMIT_IP", FAILURE_LIMIT)

# If this is True, the lockout checks to evaluate if the IP failure limit and
# the username failure limit has been reached before issuing the lockout.
LOCKOUT_BY_IP_USERNAME = get_setting("DEFENDER_LOCK_OUT_BY_IP_AND_USERNAME", False)

# if this is True, The users IP address will not get locked when
# there are too many login attempts.
DISABLE_IP_LOCKOUT = get_setting("DEFENDER_DISABLE_IP_LOCKOUT", False)

# If this is True, usernames will not get locked when
# there are too many login attempts.
DISABLE_USERNAME_LOCKOUT = get_setting("DEFENDER_DISABLE_USERNAME_LOCKOUT", False)

# use a specific username field to retrieve from login POST data
USERNAME_FORM_FIELD = get_setting("DEFENDER_USERNAME_FORM_FIELD", "username")

# see if the django app is sitting behind a reverse proxy
BEHIND_REVERSE_PROXY = get_setting("DEFENDER_BEHIND_REVERSE_PROXY", False)

# the prefix for these keys in your cache.
CACHE_PREFIX = get_setting("DEFENDER_CACHE_PREFIX", "defender")

# if the django app is behind a reverse proxy, look for the
# ip address using this HTTP header value
REVERSE_PROXY_HEADER = get_setting(
    "DEFENDER_REVERSE_PROXY_HEADER", "HTTP_X_FORWARDED_FOR"
)

try:
    # how long to wait before the bad login attempt gets forgotten. in seconds.
    COOLOFF_TIME = int(get_setting("DEFENDER_COOLOFF_TIME", 300))  # seconds
except ValueError:  # pragma: no cover
    raise Exception("DEFENDER_COOLOFF_TIME needs to be an integer")  # pragma: no cover

LOCKOUT_TEMPLATE = get_setting("DEFENDER_LOCKOUT_TEMPLATE")

ERROR_MESSAGE = gettext_lazy(
    "Please enter a correct username and password. "
    "Note that both fields are case-sensitive."
)

LOCKOUT_URL = get_setting("DEFENDER_LOCKOUT_URL")

USE_CELERY = get_setting("DEFENDER_USE_CELERY", False)

STORE_ACCESS_ATTEMPTS = get_setting("DEFENDER_STORE_ACCESS_ATTEMPTS", True)

# Used by the management command to decide how long to keep access attempt
# recods. Number is # of hours.
try:
    ACCESS_ATTEMPT_EXPIRATION = int(
        get_setting("DEFENDER_ACCESS_ATTEMPT_EXPIRATION", 24)
    )
except ValueError:  # pragma: no cover
    raise Exception(
        "DEFENDER_ACCESS_ATTEMPT_EXPIRATION" " needs to be an integer"
    )  # pragma: no cover


GET_USERNAME_FROM_REQUEST_PATH = get_setting(
    "DEFENDER_GET_USERNAME_FROM_REQUEST_PATH", "defender.utils.username_from_request"
)



from django.urls import path, re_path
from .views import block_view, unblock_ip_view, unblock_username_view

urlpatterns = [
    path("blocks/", block_view, name="defender_blocks_view"),
    re_path(
        "blocks/ip/(?P<ip_address>[A-Za-z0-9-._]+)/unblock",
        unblock_ip_view,
        name="defender_unblock_ip_view",
    ),
    path(
        "blocks/username/<path:username>/unblock",
        unblock_username_view,
        name="defender_unblock_username_view",
    ),
]



from django.db import models


class AccessAttempt(models.Model):
    """ Access Attempt log """

    user_agent = models.CharField(max_length=255,)
    ip_address = models.GenericIPAddressField(verbose_name="IP Address", null=True,)
    username = models.CharField(max_length=255, null=True,)
    http_accept = models.CharField(verbose_name="HTTP Accept", max_length=1025,)
    path_info = models.CharField(verbose_name="Path", max_length=255,)
    attempt_time = models.DateTimeField(auto_now_add=True,)
    login_valid = models.BooleanField(default=False,)

    class Meta:
        ordering = ["-attempt_time"]

    def __str__(self):
        """ unicode value for this model """
        return "{0} @ {1} | {2}".format(
            self.username, self.attempt_time, self.login_valid
        )



from django.shortcuts import render
from django.http import HttpResponseRedirect
from django.contrib.admin.views.decorators import staff_member_required
from django.urls import reverse


from .utils import get_blocked_ips, get_blocked_usernames, unblock_ip, unblock_username


@staff_member_required
def block_view(request):
    """ List the blocked IP and Usernames """
    blocked_ip_list = get_blocked_ips()
    blocked_username_list = get_blocked_usernames()

    context = {
        "blocked_ip_list": blocked_ip_list,
        "blocked_username_list": blocked_username_list,
    }
    return render(request, "defender/admin/blocks.html", context)


@staff_member_required
def unblock_ip_view(request, ip_address):
    """ upblock the given ip """
    if request.method == "POST":
        unblock_ip(ip_address)
    return HttpResponseRedirect(reverse("defender_blocks_view"))


@staff_member_required
def unblock_username_view(request, username):
    """ unblockt he given username """
    if request.method == "POST":
        unblock_username(username)
    return HttpResponseRedirect(reverse("defender_blocks_view"))



from django.db import models, migrations


class Migration(migrations.Migration):
    """ Initial migrations """

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AccessAttempt",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("user_agent", models.CharField(max_length=255)),
                (
                    "ip_address",
                    models.GenericIPAddressField(null=True, verbose_name="IP Address"),
                ),
                ("username", models.CharField(max_length=255, null=True)),
                (
                    "http_accept",
                    models.CharField(max_length=1025, verbose_name="HTTP Accept"),
                ),
                ("path_info", models.CharField(max_length=255, verbose_name="Path")),
                ("attempt_time", models.DateTimeField(auto_now_add=True)),
                ("login_valid", models.BooleanField(default=False)),
            ],
            options={"ordering": ["-attempt_time"],},
            bases=(models.Model,),
        ),
    ]



import os
from celery import Celery

PROJECT_DIR = lambda base: os.path.abspath(
    os.path.join(os.path.dirname(__file__), base).replace("\\", "/")
)


MEDIA_ROOT = PROJECT_DIR(os.path.join("media"))
MEDIA_URL = "/media/"
STATIC_ROOT = PROJECT_DIR(os.path.join("static"))
STATIC_URL = "/static/"

STATICFILES_DIRS = (PROJECT_DIR(os.path.join("media", "static")),)

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": PROJECT_DIR("defender.sb"),
    }
}


SITE_ID = 1

MIDDLEWARE = (
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "defender.middleware.FailedLoginMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware"
)

ROOT_URLCONF = "exampleapp.urls"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.admin",
    "django.contrib.staticfiles",
    "defender",
]

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
)

SECRET_KEY = os.environ.get("SECRET_KEY", "too-secret-for-test")

LOGIN_REDIRECT_URL = "/admin"

DEFENDER_LOGIN_FAILURE_LIMIT = 1
DEFENDER_COOLOFF_TIME = 60
DEFENDER_REDIS_URL = "redis://localhost:6379/1"
# don't use mock redis in unit tests, we will use real redis on CI.
DEFENDER_MOCK_REDIS = False
# Let's use custom function and strip username string from request.
DEFENDER_GET_USERNAME_FROM_REQUEST_PATH = (
    "exampleapp.utils.strip_username_from_request"
)

# Celery settings:
CELERY_ALWAYS_EAGER = True
BROKER_BACKEND = "memory"
BROKER_URL = "memory://"

# set the default Django settings module for the 'celery' program.
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "exampleapp.settings")

app = Celery("defender")

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object("django.conf:settings")
app.autodiscover_tasks(lambda: INSTALLED_APPS)

DEBUG = True

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]


from .data import store_login_attempt
from . import config


def add_login_attempt_task(
    user_agent, ip_address, username, http_accept, path_info, login_valid
):
    """ Create a record for the login attempt """
    store_login_attempt(
        user_agent, ip_address, username, http_accept, path_info, login_valid
    )

if config.USE_CELERY:
    from celery import shared_task
    add_login_attempt_task = shared_task(add_login_attempt_task)



from django.contrib import admin
from .models import AccessAttempt


class AccessAttemptAdmin(admin.ModelAdmin):
    """ Access attempt admin config """

    list_display = (
        "attempt_time",
        "ip_address",
        "user_agent",
        "username",
        "path_info",
        "login_valid",
    )

    search_fields = [
        "ip_address",
        "username",
    ]

    date_hierarchy = "attempt_time"

    fieldsets = (
        (None, {"fields": ("path_info", "login_valid")}),
        ("Meta Data", {"fields": ("user_agent", "ip_address")}),
    )


admin.site.register(AccessAttempt, AccessAttemptAdmin)



#!/usr/bin/env python
import os
from setuptools import setup, find_packages


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [
        (dirpath.replace(package + os.sep, "", 1), filenames)
        for dirpath, dirnames, filenames in os.walk(package)
        if not os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename) for filename in filenames])
    return {package: filepaths}


setup(
    name="django-defender",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="redis based Django app that locks out users after too "
    "many failed login attempts.",
    long_description="redis based Django app based on speed, that locks out"
    "users after too many failed login attempts.",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Security",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="django, cache, security, authentication, throttle, login",
    author="Ken Cochrane",
    url="https://github.com/kencochrane/django-defender",
    author_email="kencochrane@gmail.com",
    license="Apache 2",
    include_package_data=True,
    packages=find_packages(),
    package_data=get_package_data("defender"),
    python_requires='~=3.5',
    install_requires=["Django", "redis"],
    tests_require=[
        "mockredispy>=2.9.0.11,<3.0",
        "coverage",
        "celery",
        "django-redis-cache",
    ],
)



from django.urls import include, re_path
from django.conf import settings
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.conf.urls.static import static

admin.autodiscover()

urlpatterns = [
    re_path(r"^admin/", admin.site.urls),
    re_path(r"^admin/defender/", include("defender.urls")),
]


urlpatterns += staticfiles_urlpatterns()
urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



from django.dispatch import Signal

username_block = Signal()  # (providing_args=["username"])
username_unblock = Signal()  # (providing_args=["username"])
ip_block = Signal()  # (providing_args=["ip_address"])
ip_unblock = Signal()  # (providing_args=["ip_address"])


class BlockSignal:
    """
    Providing a sender is mandatory when sending signals, hence
    this empty sender class.
    """

    pass


def send_username_block_signal(username):
    username_block.send(sender=BlockSignal, username=username)


def send_username_unblock_signal(username):
    username_unblock.send(sender=BlockSignal, username=username)


def send_ip_block_signal(ip_address):
    ip_block.send(sender=BlockSignal, ip_address=ip_address)


def send_ip_unblock_signal(ip_address):
    ip_unblock.send(sender=BlockSignal, ip_address=ip_address)
